# wrapper.py
# by James Fulford

from accessors import access
from tools import ITERABLES, DICTIONARIES


class Wrapper(object):
    def __init__(self, data, splitter="__"):
        self.__data = data
        self.__splitter = splitter

    def __getattr__(self, name):
        """
        .get_{} will return a function that when called will access from self
            and pass kwargs to access

        .{} will work as normal.

        """

        if name in self.__dict__:
            return self.__dict__[name]

        if name.startswith("get_"):

            def get_ter(**kwargs):
                keypath = name[4:].split(self.__splitter)
                try:
                    return access(keypath, **kwargs)(self.__data)
                except KeyError:
                    raise AttributeError("Cannot access {}".format(keypath))

            return get_ter

        # if name.startswith("valid_"):
        #     return valid(name[6:])

        try:
            return self.__data[name]
        except KeyError:
            raise AttributeError("Cannot access {}".format(name))

    def __getitem__(self, name):
        result = access(name)(self.__data)
        if (isinstance(result, ITERABLES) or
                isinstance(result, DICTIONARIES)):
            return Wrapper(result)
        return result

    def __str__(self):
        return str(self.__data)

    def __repr__(self):
        return "{}({!r}, splitter={!r})".format(
            self.__class__.__name__, self.__data, self.__splitter
        )


if __name__ == "__main__":
    james = Wrapper({"help": {"text": "Eat something tasty"}})
    print james.get_help__text()



import logging
from accessors import *
from validators import *
from tools import flatten_list, check, partition

from tools import ITERABLES
from tools import DICTIONARIES
from tools import STRINGS

logger = logging.getLogger(__name__)


def _retrieve(data, form, *args):
    """
    Using accessors specified in form,
    retrieves all values from dictionary 'data'.
    form = {
        "value": accessor_function,
    }

    If accessor is a dict, recurses.
    If accessor is a str/unicode, splits on "/"

    Returns dictionary if succeeded. If accessor raises Exception,
    will return None (should be filtered out)
    and prints and logs a message about that entry.
    """
    entry = {}
    for tupl in form.items():
        key, accessor = tuple(tupl)  # just in case it's a list

        #
        # if accessor is a dict, recurse
        #
        if isinstance(accessor, DICTIONARIES):
            entry[key] = _retrieve(data, accessor, *args)
            if entry[key] is None:  # retrieve failed
                return None  # filter it out
        #
        # else, access the value
        #
        else:
            #
            # if value is a string, use standard accessor with / as lister
            #
            if isinstance(accessor, STRINGS):
                accessor = access(accessor.split("/"))

            #
            # try to access the value from the data
            #
            try:
                entry[key] = accessor(data, *args)
            except Exception as e:
                #
                # accessing failed. Log message
                # and return None so it gets filtered out.
                #
                msg = "({}, {}): {}\nSkipping {}.".format(
                    *(key, accessor.__name__, str(e), data)
                )
                logger.debug(msg)
                print msg
                return None
    return entry


class Nexus(object):
    def __init__(self, template):
        self.template = template  # what a valid entry looks like
        self.entries = []  # valid entries
        self.rejects = []  # all entries that fail validation
        self.packed = []  # all entries that failed unpacking

    def add(self, entries, unpackers, *args):
        """
        Adds valid entries in given format to self.entries
        Invalid entries added to self.rejects
        Un-unpackable entries added to self.packed

        Passes extra args to each getter.
        """

        def transform(entry):
            return _retrieve(entry, unpackers, *args)

        results = map(transform, entries)  # unpack

        #
        # Validate
        #
        parting = partition(lambda e: check(e, self.template) if e else e,
                            results)

        # Store results
        self.entries.extend(  # successful mappings and valid
            parting.get(True, [])
        )
        self.rejects.extend(  # successful mappings and invalid
            parting.get(False, [])
        )
        self.packed.extend(  # unsuccessful mappings
            parting.get(None, [])
        )

    def get_rows(self, columns, fill="", *args):
        """
        Returns data as rows according to given columns

        Examples given use columns:
        [
            (0, access("x", default=0, mapping=IDENTITY)),
            (2, access("y", mapping=abs))
        ]

        (Columns can also be a dictionary)
        """

        def transform(entry):
            """
            >>> transform({"x": 2.33, "y": -3.14})
            {0: 2.33, 2: 3.14}
            Returns None if an accessor raised Exception.
            """
            return _retrieve(entry, columns, *args)

        def make_row(ent, fill=fill):
            """
            Keys will be index values,
            Values will be values.
            If key not in ent, will use fill as value.

            >>> make_row({0: 2.33, 2: 3.14})
            [2.33, "", 3.14]
            """
            return map(lambda i: ent.get(i, fill),  # get ith value from ent
                       range(max(ent.keys() + 1)))  # 0 to largest col

        # Filter out formations that failed
        formed_dicts = filter(lambda a: a, map(transform, self.entries))

        # Return as rows
        return map(make_row, formed_dicts)

    def get_dictionaries(self, form, *args):
        """
        Given dictionary (form), runs accessor values on each
        of the entries stored.

        Example form is {
            "time": access("x"),
            "measures": {
                "pirates": access("y1", default=0),
                "CO 2": access("y2", mapping=lambda n: round(n, 3))
            }
        }
        """
        def transform(entry):
            return _retrieve(entry, form, *args)

        return filter(lambda r: r, map(transform, self.entries))

    def get_writeable_rows(self, columns, fill="", delimiter=",",
                           subdelimiter="|", headers=True, *args):
        """
        Returns list of delimited strings from self.get_rows()

        Lists of lists will be flattened.
        (unable to represent lists of lists in strings)
        """
        def stringify_cell(self, value):
            if isinstance(value, ITERABLES):
                value = flatten_list(value)
                value = map(self._stringify_cell, value)
                value.sort(key=lambda s: int(s) if s.isdigit() else s)
                return subdelimiter.join(value)
            else:
                return str(value).strip()

        output = self.get_rows(columns, fill=fill, *args)

        if headers:
            output = [map(lambda s: str(s[0]).strip(), self.columns)] + output

        return map(lambda r: delimiter.join(
            map(lambda v: stringify_cell(v).replace(delimiter, "|"), r)
        ), output)

    def validate(self, template, **contexts):
        """
        Using each template, which has a similar structure to each entry,
        will apply each leaf function on each entry in self.entries.
        """

        parting = partition(lambda e: check(e, template, **contexts),
                            self.entries)

        #
        # Replace entries with valid entries
        # Add invalid entries to rejects
        #
        self.entries = parting.get(True, [])
        self.rejects.extend(parting.get(False, []))


if __name__ == "__main__":
    entries = [
        {
            "x": 1,
            "y": 2,
            "james": {
                "id": 10
            },
        },
        {
            "y": 4,
            "james": {
                "id": 20
            },
            "extra_data": {
                "x": 2
            }
        },
        {
            "x": 3,
            "james": {
                "id": 30
            },
            "extra_data": {
                "x": 2
            }
        },
    ]

    unpackers = {
        "id": access(["james", "id"]),
        "x_coord": try_deeper_key("x", "extra_data", default=0),
        "y_coord": access(["y"], 0)
    }

    declaration = {
        "id": be_an(int),
        "x_coord": be_an(int),
        "y_coord": be_an(int)
    }

    print unpackers
    n = Nexus(declaration)

    n.add(entries, unpackers)
    n.validate({"id": be_unique("id")}, entries=n.entries)
    print "Entries", len(n.entries)
    print "Rejects", len(n.rejects)
    print "Packed ", len(n.packed)



# validators.py
# by James Fulford

#
# Probably the most moral Python I've ever written
# -James Fulford
#

from accessors import access
from tools import ITERABLES
from tools import ReadableFunction as validator

#
# CONTEXT DEFINITIONS
#
# 'record' context: validating against other fields in entry
#       (useful for one field less than another)
#
# 'entries' context: validating against other entries in list
#       (useful for graph references and uniqueness)
#
# 'collection' context: validating against entries in named lists (dictionary)
#       (useful for foreign references)
#


@validator
def should(*args):
    """
    Aggregates multiple tests in an 'AND' fashion (all must pass)

    If a test is a list, at least 1 must pass (OR)
        If a subtest is a list, all must pass (AND)
            If a subsubtest is a list, at least 1 must pass (OR)
    (Continues to alternate.)

    Stops list execution if
        'AND' lists: one fails
        'OR' lists: one succeeds

    >>> should(
        # ALL
        be_a(dict),  # YEP
        [   # EITHER
            [   # ALL
                have("death"),  # YEP
                lambda d: d["death"] is True  # NOPE - "AND" FAILS
            ],
            have("liberty", be_an(int)),  # YEP - OR SUCCEEDS
        ]  # YEP - AND SUCCEEDS
    )({"liberty": 22, "death": False})
    True
    """
    return _parse_tests(args, top_is_and=True)


@validator
def at_least(*args):
    """
    Aggregates multiple tests in an 'OR' fashion (1 must pass)

    If a test is a list, all must pass (AND)
        If a subtest is a list, at least 1 must pass (OR)
            If a subsubtest is a list, all must pass (AND)
    (Continues to alternate.)

    Stops list execution if
        'AND' lists: one fails
        'OR' lists: one succeeds

    >>> case(
        # EITHER
        be_a(dict),  # YEP - STOPS EXECUTION
        [   # ALL
            [   # EITHER
                have("death"),
                lambda d: d["death"] is True
            ],
            have("liberty", be_an(int)),
        ]
    )({"liberty": 22, "death": False})
    True
    """
    return _parse_tests(args, top_is_and=False)


@validator
def case(*args):
    #
    # TODO: Make it like at_least, but on an XOR basis
    # (ONE AND ONLY ONE SUCCEEDS)
    #
    pass


def _parse_tests(args, top_is_and=True):
    #
    # TODO: Be able to specify number of successes needed
    # i.e. need at least 2 tests to pass in order for this to pass
    #

    def check_that(data, **context):

        def check_and(arg, depth=0):
            if isinstance(arg, ITERABLES):
                # print ("    " * depth) + "AND({})".format(arg)
                for a in arg:
                    #
                    # If one test fails,
                    # prematurely return False
                    #
                    if not check_or(a, depth=depth + 1):
                        # lower level alternation
                        return False
                return True
            else:
                return arg(data, **context)

        def check_or(arg, depth=0):
            if isinstance(arg, ITERABLES):
                # print ("    " * depth) + "OR({})".format(arg)
                for a in arg:
                    #
                    # If one test succeeds,
                    # prematurely return True
                    #
                    if check_and(a, depth=depth + 1):
                        # lower level alternation
                        return True
                return False
            else:
                return arg(data)

        if top_is_and:
            return check_and(args, depth=0)
        else:
            return check_or(args, depth=0)

    return check_that


#
# No context validators
#


@validator
def be(*goods):
    def is_good(data):
        return data in goods
    return is_good


@validator
def be_in(good_list):
    return be(*good_list)


@validator
def be_a(*good_types):
    def good_type_or_not(data):
        return isinstance(data, tuple(good_types))
    return good_type_or_not


be_an = be_a  # makes my english feel better


@validator
def be_truthful(data):
    return bool(data)


@validator
def be_good(good_test):
    def is_test_good(data):
        return good_test(data)
    return is_test_good


@validator
def have(key, *tests):
    """
    Returns if key successfully accessed from data.
    Will call should with args, passing in accessed value.

    >>> have("liberty")({"liberty": 32})
    True

    >>> have("liberty", be_an(int))({"liberty": 32})
    True
    """
    tests = should(*tests)

    def try_access(data, **context):
        try:
            result = access(key)(data)
        except Exception:
            return False
        else:  # make sure validation errors not caught
            return tests(result, **context)

    return try_access


#
# First Contextual Validators
#

#
# TODO: Add 'entry' context validators
#


#
# 'entries' Contextual Validators
#

#
# TODO: convert these to second context validators
#

@validator
def _reference(key, min_finds, max_finds=None, flatten=True):
    """
    Returns refer, which accesses key from each item
        in second context's list

    If data is found less than min_finds times
        or more than max_finds times (if specified)
        returns False.
    Otherwise, returns True.

    If accessed value is a list, looks for datavalue in list
        (List is flattened by default)
    """
    def refer(data, entries):
        finds = 0
        for item in entries:
            acc = access(key, flatten=flatten)(item)
            if isinstance(acc, ITERABLES):
                result = data in acc
            else:
                result = data == acc
            if result:
                finds += 1
                if max_finds and finds > max_finds:
                    return False
        return finds >= min_finds
    return refer


@validator
def be_unique(key):
    """
    Returns True if value shows up only once in the list
        after accessing key from each item in first context list.

    Returns False for ALL entries with conflicts.
    """
    return _reference(key, 1, 1)


@validator
def point_to(key):
    """
    Returns True if value shows up at least once in the list
        after accessing key from each item in first context list.

    Else, False.
    """
    return _reference(key, 1)


#
# Third Context Validators
#

#
# TODO: Specify test(s) to run on foreign referred item
#
@validator
def foreign(nexi_name, key, *tests):
    """
    Returns function which checks uniqueness on third context
    """
    # test = should(*tests) if tests else be_unique
    # ^ may have to prepare_argument
    def foreign_reference(data, me, ls, nexi):
        return be_unique(key)(data, nexi[nexi_name])
    return foreign_reference


if __name__ == "__main__":
    print should(
        # ALL
        be_a(dict),
        [   # EITHER
            [   # ALL
                have("death"),
                lambda d: d["death"] is True
            ],
            have("liberty", be_an(int)),
        ]
    )({"liberty": 22, "death": False})

    @validator
    def beans(beans, **kwargs):
        def be_a_bean(data, *context):
            return data in beans
        return be_a_bean

    print should(be_a(str),
                 should(be_in(["Bean"]))
                 )





def compose(*fns):
    """
    Returns a function that evaluates it all
    """
    def comp(*args, **kwargs):
        v = fns[0](*args, **kwargs)
        for fn in fns[1:]:
            v = fn(v)
        return v
    return comp




from tools import flatten_list, pass_through, resolve_keypath
from tools import ReadableFunction as accessor
from tools import DICTIONARIES
from tools import ITERABLES


@accessor
def access(key, default=None, flatten=None,
           mapping=None, get_attr=None,
           warn=False, split="/", **kwargs):
    """Returns a function which obtains the value of *key* in
    later-passed-in data. If key is a string and split is specified, will split
    key on split value and use that list as the key path. Default: /

    If default is not None, returned function will return *default*
    if key not found in data.

    If flatten, passes flatten to flatten_list for each entry.
    The results structure is always list_flattened.

    If mapping is not falsey, will return the result of the mapping
    or the original value if fails. Accepts dictionary or function.
    (Warn is passed to pass_through)

    If get_attr is set, will access on this attribute on each entry.
    (ex: custom dict wrapping objects, stored dict in .data)

    If faulty key, will raise KeyError.
    """

    key = resolve_keypath(key, split=split)

    # Warning: recursion.
    def get_deeper(k, data, leaf=False):
        """
        Returns data at key k.

        If data is a list, gets k from each item in the list.
        If default is set, returns default value on fail
        """
        if isinstance(data, ITERABLES):
            get_deeper.non_data_lists += 1  # set before calling
            return map(lambda dp: get_deeper(k, dp, leaf=leaf), data)
        elif isinstance(data, DICTIONARIES):
            try:
                return data[k]
            except Exception:
                if default is not None and leaf:
                    return default
                else:
                    raise  # TypeError or KeyError

    def get_deep_key_from_dictionary(data, *args):
        """
        Gets deep key from data. Ignores resource.
        """
        # handling objects
        if get_attr:
            data = getattr(data, get_attr)

        # this handles depth
        get_deeper.non_data_lists = 0  # incremented in get_deeper
        for i in range(len(key)):
            k = key[i]
            data = get_deeper(k, data, leaf=(i == len(key) - 1))

        # now getting ready to return data
        if isinstance(data, ITERABLES):
            data = flatten_list(data, depth=get_deeper.non_data_lists - 2)
            if flatten:
                data = map(lambda d: flatten_list(d, depth=flatten), data)

        if mapping:
            data = pass_through(data, mapping, warn=warn, **kwargs)

        return data

    return get_deep_key_from_dictionary


@accessor
def try_deeper_key(key, deep_key, default=None, **kwargs):
    """
    Returns getter which tries to access from parent dictionary.
    If no success, tries to grab from deep_key dictionary instead.
    If that fails, uses access's default option.
    """
    def try_access(data, key=key, deep_key=deep_key, *args):
        try:
            return access(key, **kwargs)(data, *args)
        except KeyError:
            # Build new path starting from deep_key
            key = resolve_keypath(key)
            deep_key = resolve_keypath(deep_key)
            key = deep_key + key
            # Try accessing new key path
            return access(key, default=default, **kwargs)(data, *args)

    return try_access


@accessor
def always(value):
    """Always returns the same *value* irregardless of inputs"""
    def always_return(data, *args):
        return value
    return always_return


blank = always("")
blank.__doc__ = "Always returns the empty string (\"\")."



# tools.py

from inspect import getargspec


#
# These constants define what types
# invoke what type of behavior in all
# recursive functionality in this package
#

# list functionalities
ITERABLES = (
    list,
    tuple,
    # numpy.Array?
)

# dictionary functionalities
DICTIONARIES = (
    dict,
    # collections.OrderedDict?
)

# string functionalities
STRINGS = (
    str,
    unicode,
)


def partition(fn, ls):
    """
    Calls fn on each item in ls
        returns dictionary keyed on outputs of fn
        values being lists of items with that output

    >>> partition(lambda x: x ** 2, [1, 2, 3, -2])
    {1: [1], 4: [2, -2], 9: [3]}

    >>> partition(bool, [1, 0, [], "Hello"])
    {True: [1, "Hello"], False: [0, []]}
    """
    parting = {}
    for item in ls:
        r = fn(item)
        try:
            parting[r].append(item)
        except KeyError:
            parting[r] = [item]
    return parting


def resolve_keypath(keypath, split="/"):
    """
    Converts keypath into a list of keys to pass through.
    If keypath is a string, makes list by splitting on split

    >>> resolve_keypath(["a", "b"])
    ["a", "b"]

    >>> resolve_keypath("a/b")
    ["a", "b"]

    >>> resolve_keypath("a_b", split="_")
    ["a", "b"]

    """
    if isinstance(keypath, STRINGS):
        if split:
            keypath = keypath.split(split)
        else:
            keypath = [keypath]
    return keypath


def flatten_list(list_of_lists, depth=True):
    """Given a list of lists, unpacks all the items
    and puts it in one list.

    depth specifies how many layers deep we should flatten.
        by default flattens all layers

    >>> flatten_list([[1, [2], [3, 4]], [[5, [6, 7], [8]]]])
    [1, 2, 3, 4, 5, 6, 7, 8]

    >>> flatten_list([[1, [2], [3, 4]], [[5, [6, 7], [8]]]], depth=2)
    [1, 2, 3, 4, 5, [6, 7], [8]]
    """
    result = []
    for item in list_of_lists:
        if isinstance(item, ITERABLES):
            if type(depth) is int and depth - 1 >= 0:
                # flatten this level
                result.extend(flatten_list(item, depth=depth - 1))
            elif depth:
                # flatten all the way down
                result.extend(flatten_list(item))
            else:
                # preserve
                result.append(item)
        else:
            result.append(item)
    return result


def pass_through(value, mapping, warn=False, **kwargs):
    """Tries to map value through mapping. If fails, returns input.
    Mapping can be a function or a dictionary.

    If function, key words are passed down to the function.
        Exceptions are caught and printed, returns value

    If dictionary, tries to get new value by inputting the value.
        If value is not hashable (list, set, etc.), prints and returns value

    >>> pass_through(2, lambda x: x ** 2)
    4

    >>> pass_through("color", {"color": "BLUE"})
    "BLUE"

    >>> pass_through("color", {"id": 42})
    "color"

    >>> pass_through(["unhashable"], {"id": 42})
    ["unhashable"]  # also prints error to stdout

    >>> pass_through(0, lambda x: 1/x)
    0  # prints error to stdout

    """

    # Lists cannot pass_through. Instead, we pass through all items.
    # If a staggered, unflattened list of lists,
    #   will pass all entries and lower entries through
    if isinstance(value, ITERABLES):
        return map(lambda v: pass_through(v, mapping, warn=False, **kwargs),
                   value)

    if isinstance(value, DICTIONARIES):
        try:
            return mapping.get(value, value)
        except TypeError:
            msg = "pass_through: TypeError (unhashable type?)" \
                ": cannot get {} from {}".format(value, mapping)
            print msg
            # will return value later
        except KeyError:
            # this is expected behavior
            pass  # will return value later
    else:
        try:
            return mapping(value, **kwargs)
        except Exception as e:  # calling the function failed
            msg = "pass_through: {}({}) failed\n".format(
                value, mapping.__name__) + str(e)
            print msg

    # Mapping failed - return input
    return value


def check(data, template, **context):
    """
    Checks data complies with tests specified in template.
    If test fails, prints value returned and returns False
        continues to evaluate tests
    Is part of validation.
    """

    if isinstance(template, DICTIONARIES):
        # delegate to all key-value pairs
        results = []
        for key, value in template.items():
            result = check(data[key], value, **context)
            results.append(result)
            if not result:
                break
        return reduce(lambda p, q: bool(p) and bool(q), results)
    elif isinstance(template, ITERABLES):
        # delegate to all items
        results = []
        for item in data:  # data should be a list
            result = check(item, template[0], **context)
            results.append(result)
            if not result:
                break
        return reduce(lambda p, q: bool(p) and bool(q), results)
    else:
        # Leaf of template
        contexts = {"data": data}
        contexts.update(context)

        test = template
        if isinstance(test, ReadableFunction):
            test = test._resolve_to_py_function()
        try:
            # Call the validator on this piece of data
            arguments = prepare_argument(test, contexts)
            output = test(*arguments)
        except Exception as e:
            output = str(e)
            raise

        if output is not True:
            print "{}({}) failed (output: {})" \
                .format(template.__name__, data, output)
            return False
        return True


def prepare_argument(fn, contexts):
    """
    Passes only the arguments fn expects
        from contexts dicitonary
    """
    args, varargs, varkwargs, defaults = getargspec(fn)
    arguments = map(lambda a: contexts.get(a, None), args)
    return arguments


class ReadableFunction(object):
    """
    Decorator for functions.

    Acts like the function, with an eval-able .name attribute.
    If given function returns a callable when called:
        wraps callable with ReadableFunction,
        and fills in the .name attribute to hold args used in call
    Else, just returns the value.

    __str__ or __repr__ will access .name.

    >>> @ReadableFunction
    ... def sample(a, b, c=42):
    ...     def inner_func():
    ...         return a + b + c
    ...     return inner_func

    >>> str(sample(1, 42, c=73))
    'sample(1, 42, c=73)'
    """
    def __init__(self, fn, name=None):
        self.fn = fn
        if name:
            self.name = name
        else:
            self.name = fn.__name__
        self.__name__ = self.name

    def __str__(self):
        return self.name

    __repr__ = __str__

    def _resolve_to_py_function(self):
        """
        So the inspect module can work, call this to get the
            actually defined python function
        """
        if isinstance(self.fn, ReadableFunction):
            return self.fn._resolve_to_py_function()
        else:
            return self.fn

    def __call__(self, *args, **kwargs):
        result = self.fn(*args, **kwargs)

        # if a function is returned,
        # make it a validator
        # with the name describing how it was made.
        if hasattr(result, "__call__"):
            name = self.name + "("
            if args or kwargs:
                arguments = []
                if args:
                    typetype = type(type(1))
                    args = map(lambda x: x.__name__ if isinstance(x, typetype)
                               else repr(x), args)
                    arguments.extend(args)
                if kwargs:
                    arguments.extend(map(lambda kv: str(kv[0]) + "=" +
                                         repr(kv[1]), kwargs.items()))
                name += ", ".join(arguments)
            name += ")"
            # print name
            return ReadableFunction(result, name=name)
        return result



from distutils.core import setup

desc = "Declarative processing, transforming, and validating of data."

kwargs = {
    "name": "fulford.data",
    "description": desc,
    "author": "James Patrick Fulford",
    "author_email": "james.patrick.fulford@gmail.com",
    "url": "https://github.com/jamesfulford/fulford.data",
    "license": "Apache-2.0",

    "version": "0.1.0",

    "packages": ["fulforddata"],
}

setup(
    **kwargs
)
